from ftplib import FTP
from .dataset import Dataset
from pathlib import Path
from tempfile import NamedTemporaryFile
from dearpygui import dearpygui as dpg
from zosedit.constants import tempdir

class zFTP:

    def __init__(self, host, user, password):
        self.ftp = FTP(host)
        print('Attempting to connect to', host)
        self.ftp.login(user=user, passwd=password)

    def quit(self):
        self.ftp.quit()

    def list_datasets(self, search_string: str):
        files = []
        try:
            self.ftp.dir(search_string, files.append)
        except Exception as e:
            if '550' in str(e):
                return []
            print('Error listing datasets', type(e), e)
            self.error(f'Error listing datasets:\n{e}')
            return []
        datasets = [Dataset(file_) for file_ in files[1:]]
        datasets = sorted(datasets, key=lambda x: x.is_partitioned())
        return datasets

    def get_members(self, dataset: Dataset):
        members = []
        def append(line):
            members.append(line.split()[0])
        try:
            self.ftp.dir(f"'{dataset.name}(*)'", append)
        except Exception as e:
            print('Error getting members for', dataset.name)
            print(e)
        dataset._populated = True
        return members[1:] if members else []

    def download_file(self, dataset: Dataset, member:str =None):
        name = f"{dataset.name}({member})" if member else dataset.name

        raw_data = []
        # Download file
        def write(data):
            raw_data.append(data.decode('cp500'))

        try:
            self.ftp.retrbinary(f"RETR '{name}'", write)
            raw_data = ''.join(raw_data)
        except Exception as e:
            print('Error downloading', dataset)
            self.error(f'Error downloading dataset:\n{e}')
            return

        # Group data into record_length chunks
        lines = [raw_data[i:i+dataset.record_length] for i in range(0, len(raw_data), dataset.record_length)]
        path = tempdir / name
        path.write_text('\n'.join(lines))
        return path

    def mkdir(self, dataset: Dataset):
        try:
            self.ftp.mkd(f"'{dataset.name}'")
        except Exception as e:
            print('Error creating PDS', dataset)
            print(e)
            return

    def upload(self, local_path: Path, dataset: Dataset):
        data = self.read(local_path, dataset)
        try:
            with NamedTemporaryFile() as tmp:
                tmp.write(data.encode('cp500'))
                tmp.seek(0)
                self.ftp.storbinary(f"STOR '{local_path.name}'", tmp)
        except Exception as e:
            print('Error uploading', dataset)
            self.error(f'Error uploading dataset:\n{e}')
            return False
        return True

    def read(self, local_path: Path, dataset: Dataset):
        data = local_path.read_text()
        # Pad data to record_length
        lines = [line.ljust(dataset.record_length) for line in data.split('\n')]
        return ''.join(lines)

    def error(self, message):
        if dpg.does_item_exist('error'):
            dpg.delete_item('error')
        with dpg.window(label='FTP Error', tag='error'):
            dpg.add_text(message, color=(255, 0, 0))



from dearpygui import dearpygui as dpg
from zosedit.dataset import Dataset
from zosedit.constants import tempdir
from pathlib import Path


class Tab:

    def __init__(self, path: Path, id: int, dataset: Dataset):
        self.local_path = path
        self.id = id
        self.dataset = dataset
        self.dirty = False

    def mark_dirty(self):
        dpg.configure_item(self.id, label=self.dataset.name + '*')
        self.dirty = True

    def __repr__(self):
        return f"Tab({self.local_path}, {self.id})"


class Editor:

    def __init__(self, root):
        self.root = root
        self.tabs = []

    def build(self):
        with dpg.tab_bar(tag='editor_tab_bar', reorderable=True, callback=self.on_tab_changed):
            id = dpg.add_tab(label='...', closable=False)
            self.empty_tab = Tab(None, id, None)
            self.tabs.append(self.empty_tab)

        dpg.add_input_text(tag="editor", parent='win_editor', show=False,
                               multiline=True, width=-1, height=-1, callback=self.on_editor_changed)

        with dpg.handler_registry():
            dpg.add_key_press_handler(dpg.mvKey_N, callback=self.new_file_keybind)
            dpg.add_key_press_handler(dpg.mvKey_S, callback=self.save_keybind)
            dpg.add_key_press_handler(dpg.mvKey_Tab, callback=self.switch_tab_keybind)

    def hide(self):
        dpg.hide_item('editor')
        for tab in self.tabs:
            if tab is not self.empty_tab:
                dpg.hide_item(tab.id)

        self.tabs = [self.empty_tab]

    def on_editor_changed(self):
        self.get_current_tab().mark_dirty()

    def on_tab_changed(self):
        tab = dpg.get_value('editor_tab_bar')
        tab = self.get_tab_by_id(tab)
        self.switch_to_tab(tab)

    def new_file(self):
        # Callback for creating a new file
        def create_file():
            dataset_name = dpg.get_value('new_file_dataset_input')
            # record_length = dpg.get_value('new_file_record_length')
            type_ = dpg.get_value('new_file_type')
            type_ = 'PO' if type_ == 'PDS' else 'PS'

            dataset = Dataset(dataset_name)
            dataset.record_length = 80 # HACK
            dataset.type = type_
            local_path = Path(tempdir, dataset_name)
            local_path.write_text('')
            if type_ == 'PO':
                self.root.ftp.mkdir(dataset)
            else:

                self.open_file(local_path, dataset, new=True)
            dpg.delete_item('new_file_dialog')

        # Close existing dialog
        if dpg.does_item_exist('new_file_dialog'):
            dpg.delete_item('new_file_dialog')

        # Create new dialog
        w, h = 400, 100
        with dpg.window(tag='new_file_dialog', width=w, height=h, label='New'):
            dpg.add_input_text(hint='Dataset Name', tag='new_file_dataset_input', uppercase=True,
                               on_enter=True, callback=create_file)
            # dpg.add_input_int(label='Record Length', tag='new_file_record_length', default_value=80, min_value=1, max_value=32767, step=0)
            dpg.add_combo(label='Type', items=('Normal', 'PDS'), tag='new_file_type', default_value='Normal')
            with dpg.group(horizontal=True):
                dpg.add_button(label='Create', callback=create_file)
                dpg.add_button(label='Cancel', callback=lambda: dpg.delete_item('new_file_dialog'))

        # Center dialog
        vw, vh = dpg.get_viewport_width(), dpg.get_viewport_height()
        dpg.set_item_pos('new_file_dialog', (vw/2 - w/2, vh/2 - h/2))

    def save_file(self):
        tab = self.get_current_tab()
        if not tab:
            return
        if tab.dirty:
            tab.local_path.write_text(dpg.get_value('editor'), newline='')
            if not self.root.ftp.upload(tab.local_path, tab.dataset):
                return
            tab.dirty = False
            dpg.configure_item(tab.id, label=tab.dataset.name)
            current_search = dpg.get_value('explorer_search_input')
            if current_search and current_search in tab.dataset.name:
                self.root.explorer.refresh()

    def open_file(self, local_path: Path, dataset: Dataset, new=False):
        print(f'Opening {local_path}')
        if not self.get_tab_by_name(local_path):
            self.add_tab(local_path, dataset)

        self.switch_to_tab(self.get_tab_by_name(local_path))
        if new:
            self.get_current_tab().mark_dirty()

    def add_tab(self, local_path, dataset: Dataset):
        id = dpg.add_tab(label=local_path.name, closable=True, parent='editor_tab_bar')
        tab = Tab(local_path, id, dataset)
        self.tabs.append(tab)
        self.switch_to_tab(tab)

    def switch_to_tab(self, tab: Tab):
        dpg.set_value('editor_tab_bar', tab.id)
        if tab is self.empty_tab:
            dpg.hide_item('editor')
            return
        content = open(tab.local_path).read()
        lines = [line.rstrip() for line in content.split('\n')]
        dpg.set_value('editor', '\n'.join(lines))
        dpg.show_item(tab.id)
        dpg.show_item('editor')

    def cycle_tabs(self, direction: int):
        tabs = dpg.get_item_children('editor_tab_bar')[1]
        tab = dpg.get_value('editor_tab_bar')
        index = tabs.index(tab) + direction
        index = index % len(tabs)
        tab = tabs[index]
        dpg.set_value('editor_tab_bar', tab)

    def get_current_tab(self) -> Tab:
        tab = dpg.get_value('editor_tab_bar')
        return self.get_tab_by_id(tab)

    def get_tab_by_name(self, path: Path):
        matching_tabs = [tab for tab in self.tabs if tab.local_path == path]
        if len(matching_tabs) == 0:
            return None
        return matching_tabs.pop()

    def get_tab_by_id(self, id: int):
        matching_tabs = [tab for tab in self.tabs if tab.id == id]
        if len(matching_tabs) == 0:
            return None
        return matching_tabs.pop()

    def save_keybind(self):
        if dpg.is_key_down(dpg.mvKey_Control):
            self.save_file()

    def switch_tab_keybind(self):
        if dpg.is_key_down(dpg.mvKey_Control):
            self.cycle_tabs(-1 if dpg.is_key_down(dpg.mvKey_Shift) else 1)

    def new_file_keybind(self):
        if dpg.is_key_down(dpg.mvKey_Control):
            self.new_file()



from dearpygui import dearpygui as dpg
import zosedit.panels.explorer as explorer
import zosedit.panels.editor as editor

from zosedit.constants import tempdir
from zosedit.zftp import zFTP

import platform

if platform.system() == 'Windows':
    from os import startfile

class Root:

    def __init__(self):
        self.ftp = None
        self.explorer = explorer.Explorer(self)
        self.editor = editor.Editor(self)

    def start(self):
        dpg.create_context()
        dpg.create_viewport(title='z/OS Edit', resizable=True)

        with dpg.window(label="Main") as main:
            with dpg.menu_bar():
                with dpg.menu(label="File"):
                    dpg.add_menu_item(label="New", shortcut="Ctrl+N", callback=self.editor.new_file)
                    dpg.add_menu_item(label="Save", shortcut="Ctrl+S", callback=self.editor.save_file)
                    if platform.system() == 'Windows':
                        dpg.add_menu_item(label="Open Data Directory", callback=self.open_data_directory)
                with dpg.menu(label="Session"):
                    dpg.add_menu_item(label="Login", callback=self.login)
                    dpg.add_menu_item(label="Logout", callback=self.logout)

            width = dpg.get_viewport_width()
            with dpg.group(horizontal=True):
                with dpg.child_window(label="Explorer", width=width/4, height=-1, tag='win_explorer'):
                    self.explorer.build()

                with dpg.child_window(label="Editor", width=-1, height=-1, menubar=False, tag='win_editor'):
                    self.editor.build()

        self.login()

        dpg.set_primary_window(main, True)
        dpg.setup_dearpygui()
        dpg.show_viewport()
        dpg.start_dearpygui()
        dpg.destroy_context()

    def logout(self):
        self.ftp.quit()
        self.ftp = None
        self.editor.hide()
        self.explorer.hide()

    def login(self):
        def _login():
            host = dpg.get_value('settings_host_input')
            username = dpg.get_value('settings_username_input')
            password = dpg.get_value('settings_password_input')
            if self.ftp:
                dpg.set_value('login_status', 'Closing existing connection...')
                self.ftp.quit()

            dpg.set_value('login_status', f'Connecting to {host}...')
            try:
                self.ftp = zFTP(host, username, password)
            except Exception as e:
                dpg.set_value('login_status', f'Error connecting: {e}')
                return
            dpg.show_item('explorer_search_group')
            self.explorer.refresh()
            dpg.delete_item('login_dialog')

        if dpg.does_item_exist('login_dialog'):
            dpg.delete_item('login_dialog')
        w, h = 300, 150
        with dpg.window(tag='login_dialog', label='Login', width=w, height=h):
            kwargs = {'on_enter': True, 'callback': _login, 'width': -1}
            dpg.add_input_text(hint='Host', tag='settings_host_input', default_value='QAZOS205', **kwargs)
            dpg.add_input_text(hint='Username', tag='settings_username_input', uppercase=True, **kwargs)
            dpg.add_input_text(hint='Password', tag='settings_password_input', password=True, uppercase=True, **kwargs)
            with dpg.group(horizontal=True):
                bw = w // 2 - 12
                dpg.add_button(label='Login', callback=_login, width=bw)
                dpg.add_button(label='Cancel', callback=lambda: dpg.delete_item('login_dialog'), width=bw -1)
            dpg.add_text('', tag='login_status')
        vw = dpg.get_viewport_width()
        vh = dpg.get_viewport_height()
        dpg.set_item_pos('login_dialog', (vw/2-w/2, vh/2-h/2))

    def open_data_directory(self):
        startfile(tempdir)

root = Root()

def main():
    root.start()

if __name__ == '__main__':
    main()



class Dataset:
    cols = 'volume', 'unit', 'date', 'ext', 'used', 'record_format', 'record_length', 'block_size', 'type', 'name'

    def __init__(self, line):
        self.volume: str = None
        self.unit: int = None
        self.date: str = None
        self.ext: int = None
        self.used: int = None
        self.record_format: str = None
        self.record_length: int = None
        self.block_size: int = None
        self.type: str = None
        self.name: str = None
        self._populated = False

        self.line = line
        data = line.split()
        if len(data) != len(self.cols):
            self.name = data[-1].replace("'", "")
            return

        for col, value in zip(self.cols, data):
            setattr(self, col, value)

        self.name = self.name.replace("'", "")
        # self.used = int(self.used)
        # self.ext = int(self.ext)
        self.record_length = int(self.record_length)
        # self.block_size = int(self.block_size)

    def is_partitioned(self):
        return self.type == 'PO'

    def __repr__(self):
        attrs = ', '.join(f"{col}={getattr(self, col)}" for col in self.cols)
        return f"Dataset({attrs})"



import re
from dearpygui import dearpygui as dpg
from zosedit.dataset import Dataset


class Explorer:

    def __init__(self, root):
        self.root = root

    def build(self):
        with dpg.group(horizontal=True, show=False, tag='explorer_search_group'):
            dpg.add_input_text(hint='Enter a dataset', tag='explorer_search_input',
                               on_enter=True, callback=self.refresh, width=260, uppercase=True)
            dpg.add_button(label=' O ', callback=self.refresh)

    def hide(self):
        dpg.hide_item('explorer_search_group')
        if dpg.does_item_exist('results'):
            dpg.delete_item('results')

    def refresh(self):
        # Get datasets
        search = dpg.get_value('explorer_search_input')
        if not search:
            return
        if not re.match(r"'[^']+'", search):
            if '*' not in search and len(search.split('.')[-1]) < 8:
                search = f"'{search}*'"
            else:
                search = f"'{search}'"

        # Clear existing results
        if dpg.does_item_exist('results'):
            dpg.delete_item('results')

        # List results
        with dpg.child_window(label='Results', tag='results', parent='win_explorer'):

            # Search for datasets
            dpg.add_text('Searching...', tag='search_status')
            datasets = [d for d in self.root.ftp.list_datasets(search) if d.type is not None]
            dpg.set_value('search_status', f'Found {len(datasets)} dataset(s)')

            # Create buttons for each dataset
            with dpg.table(header_row=False, policy=dpg.mvTable_SizingStretchProp):
                dpg.add_table_column(label='Name')
                for dataset in datasets:
                    with dpg.table_row():
                        if dataset.is_partitioned():
                            with dpg.table_cell():
                                with dpg.collapsing_header(label=dataset.name, bullet=False) as id:
                                    with dpg.item_handler_registry() as reg:
                                        dpg.add_item_toggled_open_handler(callback=self._populate_pds(dataset, id))
                                    dpg.bind_item_handler_registry(id, reg)
                        else:
                            with dpg.table_cell():
                                self.leaf(dataset)

    def leaf(self, dataset: Dataset, member: str = None, **kwargs):
        header = dpg.add_collapsing_header(label=member or dataset.name, leaf=True, **kwargs)
        with dpg.item_handler_registry() as reg:
            dpg.add_item_clicked_handler(callback=self._open_file(dataset, member))
        dpg.bind_item_handler_registry(header, reg)

    def populate_pds(self, dataset: Dataset, id: int):
        if dataset._populated:
            return
        members = self.root.ftp.get_members(dataset)
        if not members:
            dpg.add_text('No members found', parent=id, indent=10)
            return

        for member in members:
            self.leaf(dataset=dataset, member=member, parent=id, indent=10)


    def _populate_pds(self, dataset: Dataset, parent: int):
        return lambda: self.populate_pds(dataset, parent)

    def _open_file(self, dataset: Dataset, member: str):
        def callback():
            filename = self.root.ftp.download_file(dataset, member)
            self.root.editor.open_file(filename, dataset)
        return callback
